import {ref, Ref} from "vue";

export class Byte {
    value: boolean[]

    constructor();
    constructor(value: boolean[]);
    constructor(value: number);
    constructor(value: Byte);
    constructor(value?: boolean[] | number | Byte) {
        if (value) {
            if(value instanceof Byte)
            {
                this.value = value.value.slice();
                return;
            }
            else if (typeof value === 'number') {
                if (value < 0 || value >= 256 || !Number.isInteger(value))
                    throw new Error("Invalid argument");
                this.value = Array(8).fill(false);
                for (let i = 0; i < 8; i++) {
                    this.value[i] = (value & (1 << i)) !== 0;
                }
            } else if (value.length === 8)
                this.value = value;
            else
                throw new Error("Invalid argument");
        } else
            this.value = Array(8).fill(false);
    }

    SetValue(num: number | string | Byte) {
        if(num instanceof Byte)
        {
            this.value = num.value.slice();
            return;
        }
        if (typeof num === 'string') {
            num = parseInt(num);
        }
        if (num < 0 || num >= 256)
            throw new Error("Invalid argument");
        this.value = Array(8).fill(false);
        for (let i = 0; i < 8; i++) {
            this.value[i] = (num & (1 << i)) !== 0;
        }
    }

    GetValue(format: number = 10): number | string {
        if (format === 10)
            return this.value.reduce((acc, val, index) => acc + (val ? 2 ** index : 0), 0);
        if (format === 16)
            return this.value.reduce((acc, val, index) => acc + (val ? 2 ** index : 0), 0).toString(16).padStart(2, '0');
        throw new Error("Invalid argument");
    }

    Copy(){
        return new Byte(this);
    }

    //ignore carry
    Add(b: Byte) {
        return new Byte(((this.GetValue() as number) + (b.GetValue() as number)) % 256);
    }

    //ignore carry
    AddAndAssign(b: Byte) {
        this.SetValue(((this.GetValue() as number) + (b.GetValue() as number)) % 256);
        return this;
    }

    Sub(b: Byte) {
        if (this.GetValue() < b.GetValue())
            return new Byte(((this.GetValue() as number) + 256 - (b.GetValue() as number)) % 256);
        return new Byte(((this.GetValue() as number) - (b.GetValue() as number)) % 256);
    }

    SubAndAssign(b: Byte) {
        if (this.GetValue() < b.GetValue()) {
            this.SetValue(((this.GetValue() as number) + 256 - (b.GetValue() as number)) % 256);
            return this
        }
        this.SetValue(((this.GetValue() as number) - (b.GetValue() as number)) % 256);
        return this;
    }
    And(b: Byte) {
        return new Byte(this.value.map((val, index) => val && b.value[index]));
    }
    AndAndAssign(b: Byte) {
        this.value = this.value.map((val, index) => val && b.value[index]);
        return this;
    }
    Or(b: Byte) {
        return new Byte(this.value.map((val, index) => val || b.value[index]));
    }
    OrAndAssign(b: Byte) {
        this.value = this.value.map((val, index) => val || b.value[index]);
        return this;
    }
    Xor(b: Byte) {
        return new Byte(this.value.map((val, index) => val !== b.value[index]));
    }
    XorAndAssign(b: Byte) {
        this.value = this.value.map((val, index) => val !== b.value[index]);
        return this;
    }
    Not() {
        return new Byte(this.value.map(val => !val));
    }
    NotAndAssign() {
        this.value = this.value.map(val => !val);
        return this;
    }
    Lsl(b: Byte) {
        return new Byte(this.value.slice(b.GetValue() as number).concat(Array(b.GetValue() as number).fill(false)));
    }
    LslAndAssign(b: Byte) {
        this.value = this.value.slice(b.GetValue() as number).concat(Array(b.GetValue() as number).fill(false));
        return this;
    }
    Lsr(b: Byte) {
        return new Byte(Array(b.GetValue() as number).fill(false).concat(this.value.slice(0, 8 - (b.GetValue() as number))));
    }
    LsrAndAssign(b: Byte) {
        this.value = Array(b.GetValue() as number).fill(false).concat(this.value.slice(0, 8 - (b.GetValue() as number)));
        return this;
    }
}

export class MemoryManager {
    screenMemory: Ref<Byte[][]>;
    memory: Ref<Byte[]>;

    constructor() {
        this.screenMemory = ref(Array.from({length: 16}, () => Array.from({length: 8}, () => new Byte(255))));
        this.memory = ref(Array(128).fill(new Byte(0)));
    }

    clear() {
        for (let t of this.screenMemory.value)
            t.forEach(val => val.SetValue(0));
        this.memory.value.forEach(val => val.SetValue(0));
    }

    GetVal(t:number)
    {
        if(t>=128)
            return this.memory.value[t-128];
        else
            return this.screenMemory.value[Math.floor(t/8)][t%8];
    }
    SetVal(t:number, val:Byte)
    {
        if(t>=128)
            this.memory.value[t-128] = new Byte(val);
        else
            this.screenMemory.value[Math.floor(t/8)][t%8] = new Byte(val);
    }

    SetMemSet(memorySet: Byte[]) {
        if(memorySet.length > 128)
            throw new Error("No enough space");
        for(let i = 0;i<memorySet.length;i++)
            this.memory.value[i] = memorySet[i];
    }
}